from functools import wraps
from typing import Dict

import grpc
from grpc._utilities import RpcMethodHandler

from django_grpc.signals import grpc_request_started, grpc_got_request_exception, grpc_request_finished


class SignalWrapper:
    """
    Wraps all RPC handlers to emit signal before and after each RPC
    """
    # Names of properties that can hold RPC callback
    METHOD_PROPERTIES = ('unary_unary', 'unary_stream', 'stream_unary', 'stream_stream')

    def __init__(self, server: 'grpc.Server'):
        self.server = server

    def add_generic_rpc_handlers(self, generic_rpc_handlers: tuple):
        """
        This method does the magic. It must have same interface as `grpc.Server.add_generic_rpc_handlers`
        """
        generic_rpc_handlers[0]._method_handlers = {
            key: self._replace_method_handler(method_handler)
            for key, method_handler in generic_rpc_handlers[0]._method_handlers.items()
        }
        self.server.add_generic_rpc_handlers(generic_rpc_handlers)

    def _replace_method_handler(self, method_handler: 'RpcMethodHandler') -> 'RpcMethodHandler':
        """
        Creates identical instance of RpcMethodHandler() with wrapped method handler.
        """
        kwargs = {
            prop: getattr(method_handler, prop)
            for prop in method_handler._fields
        }
        kwargs['unary_unary'] = _unary_unary(kwargs['unary_unary'])
        kwargs['unary_stream'] = _unary_stream(kwargs['unary_stream'])
        # @TODO add support for stream-unary and stream-stream methods
        # kwargs['stream_unary'] = _unary_stream(kwargs['stream_unary'])
        # kwargs['stream_stream'] = _unary_stream(kwargs['stream_stream'])

        return RpcMethodHandler(**kwargs)

    def add_registered_method_handlers(
        self,
        service_name: str,
        method_handlers: Dict[str, grpc.RpcMethodHandler],
    ):
        # do nothing; see https://github.com/grpc/grpc/issues/36683
        # and https://github.com/grpc/grpc/pull/36696
        pass

def _unary_unary(func):
    if func is None:
        return

    @wraps(func)
    def inner(*args, **kwargs):
        grpc_request_started.send(None, request=args[0], context=args[1])
        try:
            response = func(*args, **kwargs)
        except Exception as exc:
            grpc_got_request_exception.send(None, request=args[0], context=args[1], exception=exc)
            raise
        else:
            grpc_request_finished.send(None, request=args[0], context=args[1])
        return response

    return inner


def _unary_stream(func):
    if func is None:
        return

    @wraps(func)
    def inner(*args, **kwargs):
        grpc_request_started.send(None, request=args[0], context=args[1])
        try:
            for it in func(*args, **kwargs):
                yield it
        except Exception as exc:
            grpc_got_request_exception.send(None, request=args[0], context=args[1], exception=exc)
            raise
        else:
            grpc_request_finished.send(None, request=args[0], context=args[1])

    return inner



from google.protobuf.message import Message

from typing import Iterable
from django.db.models import ForeignKey, Model
from django.db.models.fields.reverse_related import ForeignObjectRel


class BaseModelSerializer:
    def __init__(self, model_class, serializers=None):
        self.model_class = model_class
        self.serializers = serializers

    def _to_dict(self, grpc_fields, instance):
        return {
            name: self._get_field_value(grpc_field, name, instance)
            for name, grpc_field in grpc_fields
        }

    def _get_field_value(self, grpc_field, name, instance):
        if instance is None:
            return None

        # Check if child serializer has special handling method
        method = getattr(self, "get_" + name, None)
        if method:
            return method(instance)

        # Default behaviour
        field_value = getattr(instance, name)
        field_meta = instance._meta.get_field(name)
        if isinstance(field_meta, ForeignObjectRel):
            return self._serialize_model_relations(grpc_field, field_value.all(), self.serializers)
        elif isinstance(field_meta, ForeignKey):
            return self.serialize_model(
                self.get_grpc_message_class(grpc_field),
                field_value,
                self.serializers
            )
        else:
            return field_value

    @classmethod
    def find_for_model(cls, instance, serializers: Iterable):
        serializer_candidates = [s for s in serializers if isinstance(instance, s.model_class)]
        return serializer_candidates[0] if len(serializer_candidates) else cls(instance.__class__, serializers)

    @classmethod
    def serialize_model(cls, message_class, instance: 'Model', serializers):
        serializer = cls.find_for_model(instance, serializers)
        serializer.serializers = serializers
        return message_class(**serializer._to_dict(message_class.DESCRIPTOR.fields_by_name.items(), instance))

    @classmethod
    def _serialize_model_relations(cls, grpc_field, related_items, serializers):
        message_class = cls.get_grpc_message_class(grpc_field)
        return [
            cls.serialize_model(message_class, it, serializers)
            for it in related_items
        ]

    @classmethod
    def get_grpc_message_class(cls, grpc_field):
        return grpc_field.message_type._concrete_class


def message_to_python(message) -> dict:
    """
    Convert fields in gRPC message to a dict
    """
    return {
        field.name: _message_value(val)
        for field, val in message.ListFields()
    }


def _message_value(val):
    """
    Check if nested values need to be deserialized
    """
    class_name = val.__class__.__name__
    # List of structures
    # Convert repeated
    # if isinstance(val, RepeatedCompositeContainer):
    if class_name == 'RepeatedCompositeContainer':
        return [
            message_to_python(it)
            for it in val
        ]
    # List of simple types
    if class_name == 'RepeatedScalarContainer':
        return list(val)

    # Convert single complex type (structure)
    if isinstance(val, Message):
        return message_to_python(val)

    # Simple type (str, int, bool)
    return val



from .base import BaseModelSerializer, message_to_python


def serialize_model(message_class, instance, serializers):
    """
    Shortcut
    """
    return BaseModelSerializer.serialize_model(message_class, instance, serializers)


def deserialize_message(message) -> dict:
    return message_to_python(message)



from django.db import reset_queries, close_old_connections
from django.dispatch import Signal


# Triggered before each RPC
# Args provided by signal: request, context
grpc_request_started = Signal()

# Triggered after each RPC
# Args provided by signal: request, context
grpc_request_finished = Signal()

# Triggered if there was an error during RPC execution
# Args provided by signal: request, context, exception
grpc_got_request_exception = Signal()

# Reset database connections between requests
grpc_request_started.connect(reset_queries)
grpc_request_started.connect(close_old_connections)
grpc_request_finished.connect(close_old_connections)

# Triggered when the server receives graceful shut down signal
grpc_shutdown = Signal()



import datetime
import asyncio

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand
from django.utils import autoreload
from django.conf import settings

from django_grpc.signals import grpc_shutdown
from django_grpc.utils import create_server, extract_handlers


class Command(BaseCommand):
    help = "Run gRPC server"
    config = getattr(settings, "GRPCSERVER", dict())

    def add_arguments(self, parser):
        parser.add_argument("--max_workers", type=int, help="Number of workers")
        parser.add_argument("--port", type=int, default=50051, help="Port number to listen")
        parser.add_argument("--autoreload", action="store_true", default=False)
        parser.add_argument(
            "--list-handlers",
            action="store_true",
            default=False,
            help="Print all registered endpoints",
        )

    def handle(self, *args, **options):
        is_async = self.config.get("async", False)
        if is_async is True:
            self._serve_async(**options)
        else:
            if options["autoreload"] is True:
                self.stdout.write("ATTENTION! Autoreload is enabled!")
                if hasattr(autoreload, "run_with_reloader"):
                    # Django 2.2. and above
                    autoreload.run_with_reloader(self._serve, **options)
                else:
                    # Before Django 2.2.
                    autoreload.main(self._serve, None, options)
            else:
                self._serve(**options)

    def _serve(self, max_workers, port, *args, **kwargs):
        """
        Run gRPC server
        """
        autoreload.raise_last_exception()
        self.stdout.write("gRPC server starting at %s" % datetime.datetime.now())

        server = create_server(max_workers, port)

        server.start()

        self.stdout.write("gRPC server is listening port %s" % port)

        if kwargs["list_handlers"] is True:
            self.stdout.write("Registered handlers:")
            for handler in extract_handlers(server):
                self.stdout.write("* %s" % handler)

        server.wait_for_termination()
        # Send shutdown signal to all connected receivers
        grpc_shutdown.send(None)

    def _serve_async(self, max_workers, port, *args, **kwargs):
        """
        Run gRPC server in async mode
        """
        self.stdout.write("gRPC async server starting  at %s" % datetime.datetime.now())

        # Coroutines to be invoked when the event loop is shutting down.
        _cleanup_coroutines = []

        server = create_server(max_workers, port)

        async def _main_routine():
            await server.start()
            self.stdout.write("gRPC async server is listening port %s" % port)

            if kwargs["list_handlers"] is True:
                self.stdout.write("Registered handlers:")
                for handler in extract_handlers(server):
                    self.stdout.write("* %s" % handler)

            await server.wait_for_termination()

        async def _graceful_shutdown():
            # Send the signal to all connected receivers on server shutdown.
            # https://github.com/gluk-w/django-grpc/issues/31
            grpc_shutdown.send(None)

        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(_main_routine())
        finally:
            loop.run_until_complete(*_cleanup_coroutines)
            loop.close()
